@@ -1,4 +1,7 @@
-__version__ = '3.1.3'
+import os
+import sys
+
+__version__ = '4.0.0'
 __spec_version__ = '1.2'
 spec_version_mapping = {
     '0.1': '<=0.4.3',
@@ -12,3 +15,5 @@ spec_reindex_mapping = {
     '1.1': True,
     '1.2': True,
 }
+
+sys.path.append(os.getcwd())
